//////////////////////////////////////////////////////////////////////////////
// deKoratorclient.cc
// -------------------
// DeKorator window decoration for KDE
// -------------------
// Autor: moty rahamim
// Email: moty.rahamim@gmail.com
// Based on Example-0.8 by David Johnson <david@usermode.org>
//////////////////////////////////////////////////////////////////////////////

#include <kconfig.h>
#include <kdeversion.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <klocale.h>

#include <qbitmap.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpainter.h>
#include <qtooltip.h>

#include <qpixmap.h>

#include <kiconeffect.h>

#include "deKoratorclient.h"

using namespace DeKorator;

// global constants

static int BUTTONSIZE = 24;
static const int DECOSIZE = 24;
static int TITLESIZE = 24;
static int FRAMESIZE = 3;









//////////////////////////////////////////////////////////////////////////////
// DeKoratorFactory Class                                                     //
//////////////////////////////////////////////////////////////////////////////

bool DeKoratorFactory::initialized_ = false;
Qt::AlignmentFlags DeKoratorFactory::titlealign_ = Qt::AlignHCenter;
bool DeKoratorFactory::colorizeFrames_ = false;
bool DeKoratorFactory::colorizeButtons_ = false;
bool DeKoratorFactory::needInit_ = false;
QString DeKoratorFactory::framesPath_ = "";
QString DeKoratorFactory::buttonsPath_ = "";
QString DeKoratorFactory::buttonsHoveType_ = "To Gray";
float DeKoratorFactory::effectsAmount_ = 5;
bool DeKoratorFactory::useMasks_ = false;
QString DeKoratorFactory::masksPath_ = "";


extern "C" KDecorationFactory* create_factory()
{
    return new DeKorator::DeKoratorFactory();
}

//////////////////////////////////////////////////////////////////////////////
// DeKoratorFactory()
// ----------------
// Constructor

DeKoratorFactory::DeKoratorFactory()
{
    readConfig();
    initialized_ = true;
    loadPixmaps();
    if ( DeKoratorFactory::colorizeFrames_ )
        colorizeDecoPixmaps();
    if ( DeKoratorFactory::colorizeButtons_ )
        colorizeButtonsPixmaps();
}

//////////////////////////////////////////////////////////////////////////////
// ~DeKoratorFactory()
// -----------------
// Destructor

DeKoratorFactory::~DeKoratorFactory() { initialized_ = false; }

//////////////////////////////////////////////////////////////////////////////
// createDecoration()
// -----------------
// Create the decoration

KDecoration* DeKoratorFactory::createDecoration( KDecorationBridge* b )
{
    return new DeKoratorClient( b, this );
}

//////////////////////////////////////////////////////////////////////////////
// reset()
// -------
// Reset the handler. Returns true if decorations need to be remade, false if
// only a repaint is necessary

bool DeKoratorFactory::reset( unsigned long changed )
{
    // read in the configuration
    initialized_ = false;
    bool confchange = readConfig();
    initialized_ = true;

    if ( confchange ||
            ( changed & ( SettingDecoration | SettingButtons | SettingBorder | SettingColors ) ) )
    {
        if ( changed & SettingColors || DeKoratorFactory::needInit_ )
        {
            loadPixmaps();
            if ( DeKoratorFactory::colorizeFrames_ )
                colorizeDecoPixmaps();
            if ( DeKoratorFactory::colorizeButtons_ )
                colorizeButtonsPixmaps();
        }
        return true;
    }
    else
    {
        resetDecorations( changed );
        return false;
    }
}

//////////////////////////////////////////////////////////////////////////////
// readConfig()
// ------------
// Read in the configuration file

bool DeKoratorFactory::readConfig()
{
    // create a config object
    KConfig config( "kwindeKoratorrc" );
    config.setGroup( "General" );

    // grab settings
    Qt::AlignmentFlags oldalign = DeKoratorFactory::titlealign_;
    QString value = config.readEntry( "TitleAlignment", "AlignHCenter" );
    if ( value == "AlignLeft" ) titlealign_ = Qt::AlignLeft;
    else if ( value == "AlignHCenter" ) titlealign_ = Qt::AlignHCenter;
    else if ( value == "AlignRight" ) titlealign_ = Qt::AlignRight;

    bool oldColorizeFrames = DeKoratorFactory::colorizeFrames_;
    DeKoratorFactory::colorizeFrames_ = config.readBoolEntry( "ColorizeFrames", "false" );

    bool oldColorizeButtons = DeKoratorFactory::colorizeButtons_;
    DeKoratorFactory::colorizeButtons_ = config.readBoolEntry( "ColorizeButtons", "false" );

    QString oldFramesPath = DeKoratorFactory::framesPath_;
    DeKoratorFactory::framesPath_ = config.readEntry( "FramesPath", "" );

    QString oldButtonsPath = DeKoratorFactory::buttonsPath_;
    DeKoratorFactory::buttonsPath_ = config.readEntry( "ButtonsPath", "" );

    QString oldMasksPath = DeKoratorFactory::masksPath_;
    DeKoratorFactory::masksPath_ = config.readEntry( "MasksPath", "" );

    DeKoratorFactory::buttonsHoveType_ = config.readEntry( "ButtonsHoverType", "To Gray" );

    DeKoratorFactory::effectsAmount_ = ( ( float ) config.readNumEntry( "EffectAmount", 5 ) / 10 );

    bool oldUseMasks = DeKoratorFactory::useMasks_;
    DeKoratorFactory::useMasks_ = config.readBoolEntry( "UseMasks", "false" );


    if ( oldColorizeButtons != DeKoratorFactory::colorizeButtons_ ||
            oldColorizeFrames != DeKoratorFactory::colorizeFrames_ ||
            oldFramesPath != DeKoratorFactory::framesPath_ ||
            oldButtonsPath != DeKoratorFactory::buttonsPath_ ||
            oldMasksPath != DeKoratorFactory::masksPath_ ||
            oldUseMasks != DeKoratorFactory::useMasks_
       )
        DeKoratorFactory::needInit_ = true;
    else
        DeKoratorFactory::needInit_ = false;

    if ( oldalign == titlealign_ &&
            oldColorizeFrames == DeKoratorFactory::colorizeFrames_ &&
            oldColorizeButtons == DeKoratorFactory::colorizeButtons_ &&
            oldFramesPath == DeKoratorFactory::framesPath_ &&
            oldButtonsPath == DeKoratorFactory::buttonsPath_ &&
            oldMasksPath == DeKoratorFactory::masksPath_ &&
            oldUseMasks == DeKoratorFactory::useMasks_ )
        return false;
    else
        return true;
}

//////////////////////////////////////////////////////////////////////////////
// loadPixmaps()
// ------------
// Read in the configuration file

void DeKoratorFactory::loadPixmaps()
{
    QString decoPixDir = DeKoratorFactory::framesPath_;
    QString btnPixDir = DeKoratorFactory::buttonsPath_;
    QString masksPixDir = DeKoratorFactory::masksPath_;

    // deco
    midTitleBg_.load( decoPixDir + "/midTitleBg.png" );
    leftFrameBg_.load( decoPixDir + "/leftFrameBg.png" );
    buttomFrameBg_.load( decoPixDir + "/buttomFrameBg.png" );
    rightFrameBg_.load( decoPixDir + "/rightFrameBg.png" );
    topLeftCornerBg_.load( decoPixDir + "/topLeftCornerBg.png" );
    topRightCornerBg_.load( decoPixDir + "/topRightCornerBg.png" );
    buttomLeftCornerBg_.load( decoPixDir + "/buttomLeftCornerBg.png" );
    buttomRightCornerBg_.load( decoPixDir + "/buttomRightCornerBg.png" );
    leftTitleBg_.load( decoPixDir + "/leftTitleBg.png" );
    rightTitleBg_.load( decoPixDir + "/rightTitleBg.png" );
    leftButtonsBg_.load( decoPixDir + "/leftButtonsBg.png" );
    rightButtonsBg_.load( decoPixDir + "/rightButtonsBg.png" );

    // buttons
    buttonHelpPix_.load( btnPixDir + "/buttonHelp.png" );
    buttonMaxPix_.load( btnPixDir + "/buttonMax.png" );
    buttonMinPix_.load( btnPixDir + "/buttonMin.png" );
    buttonClosePix_.load( btnPixDir + "/buttonClose.png" );
    buttonStickyPix_.load( btnPixDir + "/buttonSticky.png" );
    buttonStickyDownPix_.load( btnPixDir + "/buttonStickyDown.png" );
    buttonAbovePix_.load( btnPixDir + "/buttonAbove.png" );
    buttonAboveDownPix_.load( btnPixDir + "/buttonAboveDown.png" );
    ButtbonBelowPix_.load( btnPixDir + "/buttonBelow.png" );
    ButtbonBelowDownPix_.load( btnPixDir + "/buttonBelowDown.png" );
    buttonShadePix_.load( btnPixDir + "/buttonShade.png" );
    buttonShadeDownPix_.load( btnPixDir + "/buttonShadeDown.png" );

    //masks
    topLeftCornerBitmap_ = QBitmap( masksPixDir + "/topLeftCornerBitmap.png" );
    topRightCornerBitmap_ = QBitmap( masksPixDir + "/topRightCornerBitmap.png" );
    buttomLeftCornerBitmap_ = QBitmap( masksPixDir + "/buttomLeftCornerBitmap.png" );
    buttomRightCornerBitmap_ = QBitmap( masksPixDir + "/buttomRightCornerBitmap.png" );


    FRAMESIZE = buttomFrameBg_.height();
    TITLESIZE = midTitleBg_.height();
    BUTTONSIZE = TITLESIZE;
}

//////////////////////////////////////////////////////////////////////////////
// colorizeDecoPixmaps()
// ------------
//
void DeKoratorFactory::colorizeDecoPixmaps()
{
    QImage img;
    QColor col = options() ->colorGroup( KDecoration::ColorTitleBar, true ).background();

    img = midTitleBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    midTitleBg_.convertFromImage( img, 0 );

    img = leftFrameBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    leftFrameBg_.convertFromImage( img, 0 );

    img = buttomFrameBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttomFrameBg_.convertFromImage( img, 0 );

    img = rightFrameBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    rightFrameBg_.convertFromImage( img, 0 );

    img = topLeftCornerBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    topLeftCornerBg_.convertFromImage( img, 0 );

    img = topRightCornerBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    topRightCornerBg_.convertFromImage( img, 0 );

    img = buttomLeftCornerBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttomLeftCornerBg_.convertFromImage( img, 0 );

    img = buttomRightCornerBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttomRightCornerBg_.convertFromImage( img, 0 );

    img = leftTitleBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    leftTitleBg_.convertFromImage( img, 0 );

    img = rightTitleBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    rightTitleBg_.convertFromImage( img, 0 );

    img = leftButtonsBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    leftButtonsBg_.convertFromImage( img, 0 );

    img = rightButtonsBg_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    rightButtonsBg_.convertFromImage( img, 0 );
}

//////////////////////////////////////////////////////////////////////////////
// colorizeButtonsPixmaps()
// ------------
//
void DeKoratorFactory::colorizeButtonsPixmaps()
{
    QImage img;
    QColor col = options() ->colorGroup( KDecoration::ColorButtonBg, true ).background();

    img = buttonHelpPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonHelpPix_.convertFromImage( img, 0 );

    img = buttonMaxPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonMaxPix_.convertFromImage( img, 0 );

    img = buttonMinPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonMinPix_.convertFromImage( img, 0 );

    img = buttonClosePix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonClosePix_.convertFromImage( img, 0 );

    img = buttonStickyPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonStickyPix_.convertFromImage( img, 0 );

    img = buttonStickyDownPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonStickyDownPix_.convertFromImage( img, 0 );

    img = buttonAbovePix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonAbovePix_.convertFromImage( img, 0 );

    img = buttonAboveDownPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonAboveDownPix_.convertFromImage( img, 0 );

    img = ButtbonBelowPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    ButtbonBelowPix_.convertFromImage( img, 0 );

    img = ButtbonBelowDownPix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    ButtbonBelowDownPix_.convertFromImage( img, 0 );

    img = buttonShadePix_.convertToImage();
    KIconEffect::colorize( img, col, 1.0 );
    buttonShadePix_.convertFromImage( img, 0 );
}











//////////////////////////////////////////////////////////////////////////////
// DeKoratorButton Class                                                      //
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// DeKoratorButton()
// ---------------
// Constructor

DeKoratorButton::DeKoratorButton( bool isLeft, DeKoratorClient *parent, const char *name,
                              const QString& tip, ButtonType type, QPixmap *buttonPix )
        : QButton( parent->widget(), name ), client_( parent ), type_( type ),
        deco_( 0 ), lastmouse_( NoButton )
{
    isLeft_ = isLeft;
    hover_ = false;
    setBackgroundMode( NoBackground );
    setFixedSize( BUTTONSIZE, BUTTONSIZE );
    setCursor( arrowCursor );
    if ( buttonPix ) setPixmap( buttonPix );
    QToolTip::add( this, tip );
}

DeKoratorButton::~DeKoratorButton()
{
    if ( deco_ ) delete deco_;
}

//////////////////////////////////////////////////////////////////////////////
// setPixmap()
// -----------
// Set the button decoration

void DeKoratorButton::setPixmap( QPixmap *buttonPix )
{
    if ( !buttonPix ) return ; // no bitmap, probably the menu button

    //if (deco_) delete deco_;
    //   if (type_ == ButtonClose)
    //   {
    decoPix_ = *buttonPix;
    //decoPix_->setMask(*deco_);
    //   }
    //   else
    //   {
    //     deco_ = new QBitmap(DECOSIZE, DECOSIZE, bitmap, true);
    //     deco_->setMask(*deco_);
    //   }
    repaint( false );
}

//////////////////////////////////////////////////////////////////////////////
// sizeHint()
// ----------
// Return size hint

QSize DeKoratorButton::sizeHint() const
{
    return QSize( BUTTONSIZE, BUTTONSIZE );
}

//////////////////////////////////////////////////////////////////////////////
// enterEvent()
// ------------
// Mouse has entered the button

void DeKoratorButton::enterEvent( QEvent *e )
{
    // if we wanted to do mouseovers, we would keep track of it here
    QButton::enterEvent( e );
    hover_ = true;
    repaint( false );
}

//////////////////////////////////////////////////////////////////////////////
// leaveEvent()
// ------------
// Mouse has left the button

void DeKoratorButton::leaveEvent( QEvent *e )
{
    // if we wanted to do mouseovers, we would keep track of it here
    QButton::leaveEvent( e );
    hover_ = false;
    repaint( false );
}

//////////////////////////////////////////////////////////////////////////////
// mousePressEvent()
// -----------------
// Button has been pressed

void DeKoratorButton::mousePressEvent( QMouseEvent* e )
{
    lastmouse_ = e->button();

    // translate and pass on mouse event
    int button = LeftButton;
    if ( ( type_ != ButtonMax ) && ( e->button() != LeftButton ) )
    {
        button = NoButton; // middle & right buttons inappropriate
    }
    QMouseEvent me( e->type(), e->pos(), e->globalPos(),
                    button, e->state() );
    QButton::mousePressEvent( &me );
}

//////////////////////////////////////////////////////////////////////////////
// mouseReleaseEvent()
// -----------------
// Button has been released

void DeKoratorButton::mouseReleaseEvent( QMouseEvent* e )
{
    lastmouse_ = e->button();

    // translate and pass on mouse event
    int button = LeftButton;
    if ( ( type_ != ButtonMax ) && ( e->button() != LeftButton ) )
    {
        button = NoButton; // middle & right buttons inappropriate
    }
    QMouseEvent me( e->type(), e->pos(), e->globalPos(), button, e->state() );
    QButton::mouseReleaseEvent( &me );
}

//////////////////////////////////////////////////////////////////////////////
// drawButton()
// ------------
// Draw the button

void DeKoratorButton::drawButton( QPainter *painter )
{
    if ( !DeKoratorFactory::initialized() ) return ;

    QColorGroup group;
    int dx, dy;

    //painter->fillRect(rect(), Qt::red);

    if ( isLeft_ )
        painter->drawTiledPixmap( rect(), client_->decoFactory_->leftButtonsBg_ );
    else
        painter->drawTiledPixmap( rect(), client_->decoFactory_->rightButtonsBg_ );

    if ( type_ == ButtonMenu )
    {
        // we paint the mini icon (which is 16 pixels high)
        dx = ( width() - 16 ) / 2;
        dy = ( height() - 16 ) / 2;
        if ( isDown() ) { dx++; dy++; }
        painter->drawPixmap( dx, dy, client_->icon().pixmap( QIconSet::Small,
                             QIconSet::Normal ) );
    }
    else
    {
        // otherwise we paint the deco
        dx = ( width() - decoPix_.width() ) / 2;
        dy = ( height() - decoPix_.height() ) / 2;
        if ( isDown() ) { dx++; dy++; }
        if ( hover_ )
        {
            QImage img;
            QColor col = client_->decoFactory_->options() ->colorGroup( KDecoration::ColorTitleBlend, true ).background();


            img = decoPix_.convertToImage();
            if ( DeKoratorFactory::buttonsHoveType_ == "To Gray" )
                KIconEffect::toGray( img, DeKoratorFactory::effectsAmount_ );
            else if ( DeKoratorFactory::buttonsHoveType_ == "Colorize" )
                KIconEffect::colorize( img, col, DeKoratorFactory::effectsAmount_ );
            else if ( DeKoratorFactory::buttonsHoveType_ == "DeSaturate" )
                KIconEffect::deSaturate( img, DeKoratorFactory::effectsAmount_ );
            else if ( DeKoratorFactory::buttonsHoveType_ == "To Gamma" )
                KIconEffect::toGamma( img, DeKoratorFactory::effectsAmount_ );
            //KIconEffect::colorize( img, col, 1.0 );
            // midTitleBg_.convertFromImage( img, 0 );
            painter->drawPixmap( dx, dy, img );
        }
        else
            painter->drawPixmap( dx, dy, decoPix_ );
    }
    if ( client_->isSetShade() )
    {
        painter->setPen( QColor( 70, 70, 70 ) );
        painter->drawLine( 0, BUTTONSIZE - 1, BUTTONSIZE - 1, BUTTONSIZE - 1 );
    }
}











//////////////////////////////////////////////////////////////////////////////
// DeKoratorClient Class                                                      //
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// DeKoratorClient()
// ---------------
// Constructor

DeKoratorClient::DeKoratorClient( KDecorationBridge *b, KDecorationFactory *f )
        : KDecoration( b, f )
{
    decoFactory_ = ( DeKoratorFactory* ) f;
}

DeKoratorClient::~DeKoratorClient()
{
    for ( int n = 0; n < ButtonTypeCount; n++ )
    {
        if ( button[ n ] ) delete button[ n ];
    }
}

//////////////////////////////////////////////////////////////////////////////
// init()
// ------
// Actual initializer for class

void DeKoratorClient::init()
{
    createMainWidget( WResizeNoErase | WRepaintNoErase );
    widget() ->installEventFilter( this );

    // for flicker-free redraws
    widget() ->setBackgroundMode( NoBackground );

    // setup layout
    QGridLayout *mainlayout = new QGridLayout( widget(), 4, 3 ); // 4x3 grid
    QHBoxLayout *titlelayout = new QHBoxLayout();
    titlebar_ = new QSpacerItem( 1, TITLESIZE, QSizePolicy::Expanding,
                                 QSizePolicy::Fixed );

    mainlayout->setResizeMode( QLayout::FreeResize );
    //mainlayout->addRowSpacing(0, FRAMESIZE);
    mainlayout->addRowSpacing( 3, isSetShade() ? 0 : FRAMESIZE );
    mainlayout->addColSpacing( 0, FRAMESIZE );
    mainlayout->addColSpacing( 2, FRAMESIZE );

    mainlayout->addLayout( titlelayout, 1, 1 );
    if ( isPreview() )
    {
        mainlayout->addWidget(
            new QLabel( i18n( "<b><center>Dekorator preview</center></b>" ),
                        widget() ), 2, 1 );
    }
    else
    {
        mainlayout->addItem( new QSpacerItem( 0, 0 ), 2, 1 );
    }

    // the window should stretch
    mainlayout->setRowStretch( 2, 10 );
    mainlayout->setColStretch( 1, 10 );

    // setup titlebar buttons
    for ( int n = 0; n < ButtonTypeCount; n++ ) button[ n ] = 0;
    addButtons( titlelayout, options() ->titleButtonsLeft(), true );
    titlelayout->addItem( titlebar_ );
    addButtons( titlelayout, options() ->titleButtonsRight(), false );

    // connections
    connect( this, SIGNAL( keepAboveChanged( bool ) ), SLOT( keepAboveChange( bool ) ) );
    connect( this, SIGNAL( keepBelowChanged( bool ) ), SLOT( keepBelowChange( bool ) ) );
}

//////////////////////////////////////////////////////////////////////////////
// addButtons()
// ------------
// Add buttons to title layout

void DeKoratorClient::addButtons( QBoxLayout *layout, const QString& s, bool isLeft )
{
    const unsigned char * bitmap;
    QPixmap pix;
    QString tip;

    if ( s.length() > 0 )
    {
        for ( unsigned n = 0; n < s.length(); n++ )
        {
            switch ( s[ n ] )
            {
            case 'M':                                             // Menu button
                if ( !button[ ButtonMenu ] )
                {
                    button[ ButtonMenu ] =
                        new DeKoratorButton( isLeft, this, "menu", i18n( "Menu" ),
                                           ButtonMenu );
                    connect( button[ ButtonMenu ], SIGNAL( pressed() ),
                             this, SLOT( menuButtonPressed() ) );
                    layout->addWidget( button[ ButtonMenu ] );
                }
                break;

            case 'S':                                             // Sticky button
                if ( !button[ ButtonSticky ] )
                {
                    if ( isOnAllDesktops() )
                    {
                        pix = decoFactory_->buttonStickyDownPix_;
                        tip = i18n( "Un-Sticky" );
                    }
                    else
                    {
                        pix = decoFactory_->buttonStickyPix_;
                        tip = i18n( "Sticky" );
                    }
                    button[ ButtonSticky ] =
                        new DeKoratorButton( isLeft, this, "sticky", tip,
                                           ButtonSticky, &pix );
                    connect( button[ ButtonSticky ], SIGNAL( clicked() ),
                             this, SLOT( toggleOnAllDesktops() ) );
                    layout->addWidget( button[ ButtonSticky ] );
                }
                break;

            case 'H':                                             // Help button
                if ( ( !button[ ButtonHelp ] ) && providesContextHelp() )
                {
                    button[ ButtonHelp ] =
                        new DeKoratorButton( isLeft, this, "help", i18n( "Help" ),
                                           ButtonHelp, &( decoFactory_->buttonHelpPix_ ) );
                    connect( button[ ButtonHelp ], SIGNAL( clicked() ),
                             this, SLOT( showContextHelp() ) );
                    layout->addWidget( button[ ButtonHelp ] );
                }
                break;

            case 'I':                                             // Minimize button
                if ( ( !button[ ButtonMin ] ) && isMinimizable() )
                {
                    button[ ButtonMin ] =
                        new DeKoratorButton( isLeft, this, "iconify", i18n( "Minimize" ),
                                           ButtonMin, &( decoFactory_->buttonMinPix_ ) );
                    connect( button[ ButtonMin ], SIGNAL( clicked() ),
                             this, SLOT( minimize() ) );
                    layout->addWidget( button[ ButtonMin ] );
                }
                break;

            case 'A':                                             // Maximize button
                if ( ( !button[ ButtonMax ] ) && isMaximizable() )
                {
                    if ( maximizeMode() == MaximizeFull )
                    {
                        pix = decoFactory_->buttonMaxPix_;
                        tip = i18n( "Restore" );
                    }
                    else
                    {
                        pix = decoFactory_->buttonMaxPix_;
                        tip = i18n( "Maximize" );
                    }
                    button[ ButtonMax ] =
                        new DeKoratorButton( isLeft, this, "maximize", tip,
                                           ButtonMax, &pix );
                    connect( button[ ButtonMax ], SIGNAL( clicked() ),
                             this, SLOT( maxButtonPressed() ) );
                    layout->addWidget( button[ ButtonMax ] );
                }
                break;

            case 'X':                                             // Close button
                if ( ( !button[ ButtonClose ] ) && isCloseable() )
                {
                    button[ ButtonClose ] =
                        new DeKoratorButton( isLeft, this, "close", i18n( "Close" ),
                                           ButtonClose, &( decoFactory_->buttonClosePix_ ) );
                    connect( button[ ButtonClose ], SIGNAL( clicked() ),
                             this, SLOT( closeWindow() ) );
                    layout->addWidget( button[ ButtonClose ] );
                }
                break;

            case 'F':                                             // Above button
                if ( ( !button[ ButtonAbove ] ) )
                {
                    pix = decoFactory_->buttonAbovePix_;
                    button[ ButtonAbove ] =
                        new DeKoratorButton( isLeft, this, "above",
                                           i18n( "Keep Above Others" ), ButtonAbove, keepAbove() ? &( decoFactory_->buttonAboveDownPix_ ) : &( decoFactory_->buttonAbovePix_ ) );
                    connect( button[ ButtonAbove ], SIGNAL( clicked() ),
                             this, SLOT( aboveButtonPressed() ) );
                    layout->addWidget( button[ ButtonAbove ] );
                }
                break;

            case 'B':                                             // Below button
                if ( ( !button[ ButtonBelow ] ) )
                {
                    pix = decoFactory_->ButtbonBelowPix_;
                    button[ ButtonBelow ] =
                        new DeKoratorButton( isLeft, this, "below",
                                           i18n( "Keep Below Others" ), ButtonBelow, keepBelow() ? &( decoFactory_->ButtbonBelowDownPix_ ) : &( decoFactory_->ButtbonBelowPix_ ) );
                    connect( button[ ButtonBelow ], SIGNAL( clicked() ),
                             this, SLOT( belowButtonPressed() ) );
                    layout->addWidget( button[ ButtonBelow ] );
                }
                break;

            case 'L':                                             // Shade button
                if ( ( !button[ ButtonShade && isShadeable() ] ) )
                {
                    if ( isSetShade() )
                    {
                        pix = decoFactory_->buttonShadePix_;
                        tip = i18n( "Unshade" );
                    }
                    else
                    {
                        pix = decoFactory_->buttonShadePix_;
                        tip = i18n( "Shade" );
                    }
                    button[ ButtonShade ] =
                        new DeKoratorButton( isLeft, this, "shade", tip,
                                           ButtonShade, &pix );
                    connect( button[ ButtonShade ], SIGNAL( clicked() ),
                             this, SLOT( shadeButtonPressed() ) );
                    layout->addWidget( button[ ButtonShade ] );
                }
                break;

                //             case '_':   // Spacer item
                //                 layout->addSpacing( FRAMESIZE );
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// activeChange()
// --------------
// window active state has changed

void DeKoratorClient::activeChange()
{
    for ( int n = 0; n < ButtonTypeCount; n++ )
        if ( button[ n ] ) button[ n ] ->reset();
    widget() ->repaint( false );
}

//////////////////////////////////////////////////////////////////////////////
// captionChange()
// ---------------
// The title has changed

void DeKoratorClient::captionChange()
{
    widget() ->repaint( titlebar_->geometry(), false );
}

//////////////////////////////////////////////////////////////////////////////
// desktopChange()
// ---------------
// Called when desktop/sticky changes

void DeKoratorClient::desktopChange()
{
    bool d = isOnAllDesktops();
    if ( button[ ButtonSticky ] )
    {
        button[ ButtonSticky ] ->setPixmap( d ? &( decoFactory_->buttonStickyDownPix_ ) : &( decoFactory_->buttonStickyPix_ ) );
        QToolTip::remove( button[ ButtonSticky ] );
        QToolTip::add( button[ ButtonSticky ], d ? i18n( "Sticky" ) : i18n( "Un-Sticky" ) );
    }
}

//////////////////////////////////////////////////////////////////////////////
// iconChange()
// ------------
// The title has changed

void DeKoratorClient::iconChange()
{
    if ( button[ ButtonMenu ] )
    {
        button[ ButtonMenu ] ->setPixmap( 0 );
        button[ ButtonMenu ] ->repaint( false );
    }
}

//////////////////////////////////////////////////////////////////////////////
// maximizeChange()
// ----------------
// Maximized state has changed

void DeKoratorClient::maximizeChange()
{
    bool m = ( maximizeMode() == MaximizeFull );
    if ( button[ ButtonMax ] )
    {
        button[ ButtonMax ] ->setPixmap( &( decoFactory_->buttonMaxPix_ ) );
        QToolTip::remove( button[ ButtonMax ] );
        QToolTip::add( button[ ButtonMax ], m ? i18n( "Restore" ) : i18n( "Maximize" ) );
    }
}

//////////////////////////////////////////////////////////////////////////////
// shadeChange()
// -------------
// Called when window shading changes

void DeKoratorClient::shadeChange()
{
    //FIXME
    bool s = isSetShade();
    if ( button[ ButtonShade ] )
    {
        button[ ButtonShade ] ->setPixmap( s ? &( decoFactory_->buttonShadeDownPix_ ) : &( decoFactory_->buttonShadePix_ ) );
        QToolTip::remove( button[ ButtonShade ] );
        QToolTip::add( button[ ButtonShade ], s ? i18n( "Unshade" ) : i18n( "Shade" ) );
    }

    for ( int n = 0; n < ButtonTypeCount; n++ )
        if ( button[ n ] ) button[ n ] ->reset();
    widget() ->repaint( false );
}

//////////////////////////////////////////////////////////////////////////////
// keepAboveChange()
// ------------
// The above state has changed

void DeKoratorClient::keepAboveChange( bool a )
{
    if ( button[ ButtonAbove ] )
    {
        button[ ButtonAbove ] ->setPixmap( a ? &( decoFactory_->buttonAboveDownPix_ ) : &( decoFactory_->buttonAbovePix_ ) );
    }
}

//////////////////////////////////////////////////////////////////////////////
// keepBelowChange()
// ------------
// The below state has changed

void DeKoratorClient::keepBelowChange( bool b )
{
    if ( button[ ButtonBelow ] )
    {
        button[ ButtonBelow ] ->setPixmap( b ? &( decoFactory_->ButtbonBelowDownPix_ ) : &( decoFactory_->ButtbonBelowPix_ ) );
    }
}

//////////////////////////////////////////////////////////////////////////////
// borders()
// ----------
// Get the size of the borders

void DeKoratorClient::borders( int &l, int &r, int &t, int &b ) const
{
    l = r = FRAMESIZE;
    t = TITLESIZE /*+ FRAMESIZE*/;
    b = isSetShade() ? 0 : FRAMESIZE;
}

//////////////////////////////////////////////////////////////////////////////
// resize()
// --------
// Called to resize the window

void DeKoratorClient::resize( const QSize &size )
{
    widget() ->resize( size );
}

//////////////////////////////////////////////////////////////////////////////
// minimumSize()
// -------------
// Return the minimum allowable size for this window

QSize DeKoratorClient::minimumSize() const
{
    return widget() ->minimumSize();
}

//////////////////////////////////////////////////////////////////////////////
// mousePosition()
// ---------------
// Return logical mouse position

KDecoration::Position DeKoratorClient::mousePosition( const QPoint &point ) const
{
    const int corner = 24;
    Position pos;

    if ( point.y() <= FRAMESIZE )
    {
        // inside top frame
        if ( point.x() <= corner ) pos = PositionTopLeft;
        else if ( point.x() >= ( width() - corner ) ) pos = PositionTopRight;
        else pos = PositionTop;
    }
    else if ( point.y() >= ( height() - FRAMESIZE * 2 ) )
    {
        // inside handle
        if ( point.x() <= corner ) pos = PositionBottomLeft;
        else if ( point.x() >= ( width() - corner ) ) pos = PositionBottomRight;
        else pos = PositionBottom;
    }
    else if ( point.x() <= FRAMESIZE )
    {
        // on left frame
        if ( point.y() <= corner ) pos = PositionTopLeft;
        else if ( point.y() >= ( height() - corner ) ) pos = PositionBottomLeft;
        else pos = PositionLeft;
    }
    else if ( point.x() >= width() - FRAMESIZE )
    {
        // on right frame
        if ( point.y() <= corner ) pos = PositionTopRight;
        else if ( point.y() >= ( height() - corner ) ) pos = PositionBottomRight;
        else pos = PositionRight;
    }
    else
    {
        // inside the frame
        pos = PositionCenter;
    }
    return pos;
}

//////////////////////////////////////////////////////////////////////////////
// eventFilter()
// -------------
// Event filter

bool DeKoratorClient::eventFilter( QObject *obj, QEvent *e )
{
    if ( obj != widget() ) return false;

    switch ( e->type() )
    {
    case QEvent::MouseButtonDblClick:
        {
            mouseDoubleClickEvent( static_cast<QMouseEvent *>( e ) );
            return true;
        }
    case QEvent::MouseButtonPress:
        {
            processMousePressEvent( static_cast<QMouseEvent *>( e ) );
            if ( DeKoratorFactory::useMasks_ )
                doShape();
            return true;
        }
    case QEvent::Paint:
        {
            paintEvent( static_cast<QPaintEvent *>( e ) );
            return true;
        }
    case QEvent::Resize:
        {
            resizeEvent( static_cast<QResizeEvent *>( e ) );
            return true;
        }
    case QEvent::Show:
        {
            showEvent( static_cast<QShowEvent *>( e ) );
            return true;
        }
    default:
        {
            return false;
        }
    }

    return false;
}

//////////////////////////////////////////////////////////////////////////////
// mouseDoubleClickEvent()
// -----------------------
// Doubleclick on title

void DeKoratorClient::mouseDoubleClickEvent( QMouseEvent *e )
{
    if ( titlebar_->geometry().contains( e->pos() ) ) titlebarDblClickOperation();
}

//////////////////////////////////////////////////////////////////////////////
// paintEvent()
// ------------
// Repaint the window

void DeKoratorClient::paintEvent( QPaintEvent* )
{
    if ( !DeKoratorFactory::initialized() ) return ;

    QPainter painter( widget() );
    int tx, ty, tw, th;


    // draw the titlebar

    QRect title( titlebar_->geometry() );
    title.rect( &tx, &ty, &tw, &th );
    QRect rect;

    // topLeftCorner
    rect.setRect( 0, 0, FRAMESIZE, TITLESIZE );
    painter.drawTiledPixmap( rect, decoFactory_->topLeftCornerBg_ );

    //leftTitle
    rect.setRect( tx, 0, decoFactory_->leftTitleBg_.width(), TITLESIZE );
    painter.drawTiledPixmap( rect, decoFactory_->leftTitleBg_ );

    //midTitle
    rect.setRect( tx + decoFactory_->leftTitleBg_.width(), 0, tw - ( decoFactory_->rightTitleBg_.width() + decoFactory_->leftTitleBg_.width() ), th );
    painter.drawTiledPixmap( rect, decoFactory_->leftTitleBg_ );
    painter.drawTiledPixmap( rect, decoFactory_->midTitleBg_ );

    //rightTitle
    rect.setRect( tx + tw - decoFactory_->rightTitleBg_.width(), 0, decoFactory_->rightTitleBg_.width(), th );
    painter.drawTiledPixmap( rect, decoFactory_->rightTitleBg_ );

    //topRightCorner
    rect.setRect( widget() ->width() - FRAMESIZE, 0, FRAMESIZE, TITLESIZE );
    painter.drawTiledPixmap( rect, decoFactory_->topRightCornerBg_ );


    // draw title text
    painter.setFont( options() ->font( isActive(), false ) );
    painter.setPen( options() ->color( KDecoration::ColorFont, isActive() ) );
    painter.drawText( tx + FRAMESIZE, ty,
                      tw - FRAMESIZE * 2, th,
                      DeKoratorFactory::titleAlign() | AlignVCenter,
                      caption() );

    // draw frame

    int w = width() , h = height();

    //leftFrame
    rect.setRect( 0, TITLESIZE, FRAMESIZE, h - TITLESIZE - FRAMESIZE );
    painter.drawTiledPixmap( rect, decoFactory_->leftFrameBg_ );

    if ( ! isSetShade() )
    {
        //buttomLeftCorner
        rect.setRect( 0, h - FRAMESIZE, FRAMESIZE, FRAMESIZE );
        painter.drawTiledPixmap( rect, decoFactory_->buttomLeftCornerBg_ );

        //buttomFrame
        rect.setRect( 0 + FRAMESIZE, h - FRAMESIZE, w - ( FRAMESIZE * 2 ), FRAMESIZE );
        painter.drawTiledPixmap( rect, decoFactory_->buttomFrameBg_ );

        //buttomRightCorner
        rect.setRect( w - FRAMESIZE, h - FRAMESIZE, FRAMESIZE, FRAMESIZE );
        painter.drawTiledPixmap( rect, decoFactory_->buttomRightCornerBg_ );
    }
    else
    {
        painter.setPen( QColor( 70, 70, 70 ) );
        painter.drawLine( 0 , h - 1, w, h - 1 );
    }

    //rightFrame
    rect.setRect( w - FRAMESIZE, TITLESIZE, FRAMESIZE, h - TITLESIZE - FRAMESIZE );
    painter.drawTiledPixmap( rect, decoFactory_->rightFrameBg_ );
}

//////////////////////////////////////////////////////////////////////////////
// resizeEvent()
// -------------
// Window is being resized

void DeKoratorClient::resizeEvent( QResizeEvent * )
{
    if ( widget() ->isShown() )
    {
        QRegion region = widget() ->rect();
        region = region.subtract( titlebar_->geometry() );
        widget() ->erase( region );
    }
    if ( DeKoratorFactory::useMasks_ )
        doShape();
}

//////////////////////////////////////////////////////////////////////////////
// showEvent()
// -----------
// Window is being shown

void DeKoratorClient::showEvent( QShowEvent * )
{
    widget() ->repaint();
    if ( DeKoratorFactory::useMasks_ )
        doShape();
}

//////////////////////////////////////////////////////////////////////////////
// maxButtonPressed()
// -----------------
// Max button was pressed

void DeKoratorClient::maxButtonPressed()
{
    if ( button[ ButtonMax ] )
    {
#if KDE_IS_VERSION(3, 3, 0)
        maximize( button[ ButtonMax ] ->lastMousePress() );
#else
        switch ( button[ ButtonMax ] ->lastMousePress() )
        {
        case MidButton:
            maximize( maximizeMode() ^ MaximizeVertical );
            break;
        case RightButton:
            maximize( maximizeMode() ^ MaximizeHorizontal );
            break;
        default:
            ( maximizeMode() == MaximizeFull ) ? maximize( MaximizeRestore )
            : maximize( MaximizeFull );
        }
#endif

    }
}

//////////////////////////////////////////////////////////////////////////////
// shadeButtonPressed()
// -----------------
// Shade button was pressed

void DeKoratorClient::shadeButtonPressed()
{
    if ( button[ ButtonShade ] )
    {
        setShade( !isSetShade() );
    }
}

//////////////////////////////////////////////////////////////////////////////
// aboveButtonPressed()
// -----------------
// Above button was pressed

void DeKoratorClient::aboveButtonPressed()
{
    if ( button[ ButtonAbove ] )
    {
        setKeepAbove( !keepAbove() );
    }
}

//////////////////////////////////////////////////////////////////////////////
// belowButtonPressed()
// -----------------
// Below button was pressed

void DeKoratorClient::belowButtonPressed()
{
    if ( button[ ButtonBelow ] )
    {
        setKeepBelow( !keepBelow() );
    }
}

//////////////////////////////////////////////////////////////////////////////
// menuButtonPressed()
// -------------------
// Menu button was pressed (popup the menu)

void DeKoratorClient::menuButtonPressed()
{
    if ( button[ ButtonMenu ] )
    {
        QPoint p( button[ ButtonMenu ] ->rect().bottomLeft().x(),
                  button[ ButtonMenu ] ->rect().bottomLeft().y() );
        KDecorationFactory* f = factory();
        showWindowMenu( button[ ButtonMenu ] ->mapToGlobal( p ) );
        if ( !f->exists( this ) ) return ; // decoration was destroyed
        button[ ButtonMenu ] ->setDown( false );
    }
}

//////////////////////////////////////////////////////////////////////////////
// doShape()
// -------------------
//

void DeKoratorClient::doShape()
{
    QRegion mask( 0, 0, width(), height() );
    // QRegion m(width()-FRAMESIZE, 0, FRAMESIZE, TITLESIZE );
    // m.subtract((QRegion( decoFactory_->topRightCornerBitmap_ ) );

    // top left
    QRegion m = QRegion( decoFactory_->topLeftCornerBitmap_ );
    mask -= QRegion( m );

    //top right
    m = QRegion( decoFactory_->topRightCornerBitmap_ );
    m.translate( width() - decoFactory_->topRightCornerBitmap_.width(), 0 );
    mask -= QRegion( m );

    //buttom left
    m = QRegion( decoFactory_->buttomLeftCornerBitmap_ );
    m.translate( 0, height() - decoFactory_->buttomLeftCornerBitmap_.height() );
    mask -= QRegion( m );

    //buttom right
    m = QRegion( decoFactory_->buttomRightCornerBitmap_ );
    m.translate( width() - decoFactory_->buttomRightCornerBitmap_.width(), height() - decoFactory_->topRightCornerBitmap_.height() );
    mask -= QRegion( m );

    setMask( mask );


    // Remove top-right corner.
    //     mask -= QRegion( r - 5, 0, 5, 1 );
    //     mask -= QRegion( r - 3, 1, 3, 1 );
    //     mask -= QRegion( r - 2, 2, 2, 1 );
    //     mask -= QRegion( r - 1, 3, 1, 2 );


    //     // Remove bottom-left corner.
    //     mask -= QRegion( 0, b - 0, 5, 1 );
    //     mask -= QRegion( 0, b - 1, 3, 1 );
    //     mask -= QRegion( 0, b - 2, 2, 1 );
    //     mask -= QRegion( 0, b - 4, 1, 2 );
    //
    //     // Remove bottom-right corner.
    //     mask -= QRegion( r - 5, b - 0, 5, 1 );
    //     mask -= QRegion( r - 3, b - 1, 3, 1 );
    //     mask -= QRegion( r - 2, b - 2, 2, 1 );
    //     mask -= QRegion( r - 1, b - 4, 1, 2 );

    //     setMask( mask );
}


#include "deKoratorclient.moc"
